@@ -1,7 +1,7 @@
-from collections import namedtuple
 import six
-from six import StringIO, BytesIO
 from six.moves import queue as Queue
+from collections import defaultdict
+import json
 import os.path
 import re
 from functools import partial
@@ -11,15 +11,32 @@ import botocore
 from glob import glob
 import google.cloud.exceptions
 from google.cloud.storage import Client
-import gzip
 import tenacity
+from tqdm import tqdm
 
-from .lib import mkdir
+from . import compression
+from .lib import mkdir, extract_path
 from .threaded_queue import ThreadedQueue
-from .connectionpools import S3ConnectionPool, GCloudConnectionPool
+from .connectionpools import S3ConnectionPool, GCloudBucketPool
 
-S3_POOL = S3ConnectionPool()
-GC_POOL = GCloudConnectionPool()
+# This is just to support pooling by bucket
+class keydefaultdict(defaultdict):
+    def __missing__(self, key):
+        if self.default_factory is None:
+            raise KeyError( key )
+        else:
+            ret = self[key] = self.default_factory(key)
+            return ret
+
+S3_POOL = None
+GC_POOL = None
+def reset_connection_pools():
+    global S3_POOL
+    global GC_POOL
+    S3_POOL = S3ConnectionPool()
+    GC_POOL = keydefaultdict(lambda bucket_name: GCloudBucketPool(bucket_name))
+
+reset_connection_pools()
 
 retry = tenacity.retry(
     reraise=True, 
@@ -27,6 +44,144 @@ retry = tenacity.retry(
     wait=tenacity.wait_full_jitter(0.5, 60.0),
 )
 
+DEFAULT_THREADS = 20
+
+class SimpleStorage(object):
+    """
+    Access files stored in Google Storage (gs), Amazon S3 (s3), 
+    or the local Filesystem (file).
+
+    e.g. with Storage('gs://bucket/dataset/layer') as stor:
+            files = stor.get_file('filename')
+
+    Required:
+        layer_path (str): A protocol prefixed path of the above format.
+            Accepts s3:// gs:// and file://. File paths are absolute.
+
+    Optional:
+        n_threads (int:20): number of threads to use downloading and uplaoding.
+            If 0, execution will be on the main python thread.
+        progress (bool:false): Show a tqdm progress bar for multiple 
+            uploads and downloads.
+    """
+    def __init__(self, layer_path, progress=False):
+
+        self.progress = progress
+
+        self._layer_path = layer_path
+        self._path = extract_path(layer_path)
+        
+        if self._path.protocol == 'file':
+            self._interface_cls = FileInterface
+        elif self._path.protocol == 'gs':
+            self._interface_cls = GoogleCloudStorageInterface
+        elif self._path.protocol == 's3':
+            self._interface_cls = S3Interface
+
+        self._interface = self._interface_cls(self._path)
+
+    @property
+    def layer_path(self):
+        return self._layer_path
+
+    def get_path_to_file(self, file_path):
+        return os.path.join(self._layer_path, file_path)
+
+    def put_file(self, file_path, content, content_type=None, compress=None, cache_control=None):
+        """ 
+        Args:
+            filename (string): it can contains folders
+            content (string): binary data to save
+        """
+        return self.put_files([ (file_path, content) ], 
+            content_type=content_type, 
+            compress=compress, 
+            cache_control=cache_control, 
+            block=False
+        )
+
+    def put_files(self, files, content_type=None, compress=None, cache_control=None, block=True):
+        """
+        Put lots of files at once and get a nice progress bar. It'll also wait
+        for the upload to complete, just like get_files.
+
+        Required:
+            files: [ (filepath, content), .... ]
+        """
+        for path, content in files:
+            content = compression.compress(content, method=compress)
+            self._interface.put_file(path, content, content_type, compress, cache_control=cache_control)
+        return self
+
+    def exists(self, file_path):
+        """Test if a single file exists. Returns boolean."""
+        return self._interface.exists(file_path)
+
+    def files_exist(self, file_paths):
+        """
+        Threaded exists for all file paths. 
+
+        file_paths: (list) file paths to test for existence
+
+        Returns: { filepath: bool }
+        """
+        results = {}
+        for path in file_paths:
+            results[path] = self._interface.exists(path)
+        return results
+
+    def get_json(self, file_path):
+        content = self.get_file(file_path)
+        if content is None:
+            return None
+        return json.loads(content.decode('utf8'))
+
+    def get_file(self, file_path):
+        content, encoding = self._interface.get_file(file_path)
+        content = compression.decompress(content, encoding, filename=file_path)
+        return content
+
+    def delete_file(self, file_path):
+        self._interface.delete_file(file_path)
+
+    def delete_files(self, file_paths):
+        for path in file_paths:
+            self._interface.delete_file(path)
+        return self
+
+    def list_files(self, prefix="", flat=False):
+        """
+        List the files in the layer with the given prefix. 
+
+        flat means only generate one level of a directory,
+        while non-flat means generate all file paths with that 
+        prefix.
+
+        Here's how flat=True handles different senarios:
+            1. partial directory name prefix = 'bigarr'
+                - lists the '' directory and filters on key 'bigarr'
+            2. full directory name prefix = 'bigarray'
+                - Same as (1), but using key 'bigarray'
+            3. full directory name + "/" prefix = 'bigarray/'
+                - Lists the 'bigarray' directory
+            4. partial file name prefix = 'bigarray/chunk_'
+                - Lists the 'bigarray/' directory and filters on 'chunk_'
+        
+        Return: generated sequence of file paths relative to layer_path
+        """
+
+        for f in self._interface.list_files(prefix, flat):
+            yield f
+
+    def __del__(self):
+        self._interface.release_connection()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exception_type, exception_value, traceback):
+        self._interface.release_connection()
+
 class Storage(ThreadedQueue):
     """
     Access files stored in Google Storage (gs), Amazon S3 (s3), 
@@ -45,17 +200,12 @@ class Storage(ThreadedQueue):
         progress (bool:false): Show a tqdm progress bar for multiple 
             uploads and downloads.
     """
-    gzip_magic_numbers = [ 0x1f, 0x8b ]
-    path_regex = re.compile(r'^(gs|file|s3)://(/?.*?)/(.*/)?([^//]+)/([^//]+)/?$')
-    ExtractedPath = namedtuple('ExtractedPath',
-        ['protocol','bucket_name','dataset_path','dataset_name','layer_name'])
-
     def __init__(self, layer_path, n_threads=20, progress=False):
 
         self.progress = progress
 
         self._layer_path = layer_path
-        self._path = self.extract_path(layer_path)
+        self._path = extract_path(layer_path)
         
         if self._path.protocol == 'file':
             self._interface_cls = FileInterface
@@ -85,23 +235,20 @@ class Storage(ThreadedQueue):
     def get_path_to_file(self, file_path):
         return os.path.join(self._layer_path, file_path)
 
-    @classmethod
-    def extract_path(cls, layer_path):
-        match = cls.path_regex.match(layer_path)
-        if not match:
-            return None
-        else:
-            return cls.ExtractedPath(*match.groups())
-
-    def put_file(self, file_path, content, content_type=None, compress=False):
+    def put_file(self, file_path, content, content_type=None, compress=None, cache_control=None):
         """ 
         Args:
             filename (string): it can contains folders
             content (string): binary data to save
         """
-        return self.put_files([ (file_path, content) ], content_type, compress, block=False)
+        return self.put_files([ (file_path, content) ], 
+            content_type=content_type, 
+            compress=compress, 
+            cache_control=cache_control, 
+            block=False
+        )
 
-    def put_files(self, files, content_type=None, compress=False, block=True):
+    def put_files(self, files, content_type=None, compress=None, cache_control=None, block=True):
         """
         Put lots of files at once and get a nice progress bar. It'll also wait
         for the upload to complete, just like get_files.
@@ -110,12 +257,10 @@ class Storage(ThreadedQueue):
             files: [ (filepath, content), .... ]
         """
         def base_uploadfn(path, content, interface):
-            interface.put_file(path, content, content_type, compress)
+            interface.put_file(path, content, content_type, compress, cache_control=cache_control)
 
         for path, content in files:
-            if compress:
-                content = self._compress(content)
-
+            content = compression.compress(content, method=compress)
             uploadfn = partial(base_uploadfn, path, content)
 
             if len(self._threads):
@@ -152,16 +297,20 @@ class Storage(ThreadedQueue):
             else:
                 exist_thunk(path, self._interface)
 
-        self.wait()
+        desc = 'Existence Testing' if self.progress else None
+        self.wait(desc)
 
         return results
 
-    def get_file(self, file_path):
-        # Create get_files does uses threading to speed up downloading
+    def get_json(self, file_path):
+        content = self.get_file(file_path)
+        if content is None:
+            return None
+        return json.loads(content.decode('utf8'))
 
-        content, decompress = self._interface.get_file(file_path)
-        if content and decompress != False:
-            content = self._maybe_uncompress(content)
+    def get_file(self, file_path):
+        content, encoding = self._interface.get_file(file_path)
+        content = compression.decompress(content, encoding, filename=file_path)
         return content
 
     def get_files(self, file_paths):
@@ -178,11 +327,12 @@ class Storage(ThreadedQueue):
                 result = interface.get_file(path)
             except Exception as err:
                 error = err
-                print(err)
+                # important to print immediately because 
+                # errors are collected at the end
+                print(err) 
             
-            content, decompress = result
-            if content and decompress:
-                content = self._maybe_uncompress(content)
+            content, encoding = result
+            content = compression.decompress(content, encoding)
 
             results.append({
                 "filename": path,
@@ -213,31 +363,21 @@ class Storage(ThreadedQueue):
 
         return self
 
-    def _maybe_uncompress(self, content):
-        """ Uncompression is applied if the first to bytes matches with
-            the gzip magic numbers. 
-            There is once chance in 65536 that a file that is not gzipped will
-            be ungzipped. That's why is better to set uncompress to False in
-            get file.
-        """
-        first_two_bytes = [ byte for byte in bytearray(content)[:2] ]
-        if first_two_bytes == self.gzip_magic_numbers:
-            return self._uncompress(content)
-        return content
+    def delete_files(self, file_paths):
 
-    @staticmethod
-    def _compress(content):
-        stringio = BytesIO()
-        gzip_obj = gzip.GzipFile(mode='wb', fileobj=stringio)
-        gzip_obj.write(content)
-        gzip_obj.close()
-        return stringio.getvalue()
+        def thunk_delete(path, interface):
+            interface.delete_file(path)
 
-    @staticmethod
-    def _uncompress(content):
-        stringio = BytesIO(content)
-        with gzip.GzipFile(mode='rb', fileobj=stringio) as gfile:
-            return gfile.read()
+        for path in file_paths:
+            if len(self._threads):
+                self.put(partial(thunk_delete, path))
+            else:
+                thunk_delete(path, self._interface)
+
+        desc = 'Deleting' if self.progress else None
+        self.wait(desc)
+
+        return self
 
     def list_files(self, prefix="", flat=False):
         """
@@ -277,14 +417,17 @@ class FileInterface(object):
 
     def get_path_to_file(self, file_path):
         
-        clean = filter(None,[self._path.bucket_name,
-                             self._path.dataset_path,
-                             self._path.dataset_name,
-                             self._path.layer_name,
-                             file_path])
-        return  os.path.join(*clean)
+        clean = filter(None,[
+          self._path.bucket, 
+          self._path.intermediate_path,                             
+          self._path.dataset,
+          self._path.layer,
+          file_path
+        ])
 
-    def put_file(self, file_path, content, content_type, compress):
+        return os.path.join(*clean)
+
+    def put_file(self, file_path, content, content_type, compress, cache_control=None):
         path = self.get_path_to_file(file_path)
         mkdir(os.path.dirname(path))
 
@@ -312,13 +455,15 @@ class FileInterface(object):
             
         if compressed:
             path += '.gz'
+
+        encoding = 'gzip' if compressed else None
         
         try:
             with open(path, 'rb') as f:
                 data = f.read()
-            return data, compressed
+            return data, encoding
         except IOError:
-            return None, False
+            return None, encoding
 
     def exists(self, file_path):
         path = self.get_path_to_file(file_path)
@@ -380,22 +525,25 @@ class GoogleCloudStorageInterface(object):
     def __init__(self, path):
         global GC_POOL
         self._path = path
-        self._client = GC_POOL.get_connection()
-        self._bucket = self._client.get_bucket(self._path.bucket_name)
+        self._bucket = GC_POOL[path.bucket].get_connection()
 
     def get_path_to_file(self, file_path):
-        clean = filter(None,[self._path.dataset_path,
-                             self._path.dataset_name,
-                             self._path.layer_name,
-                             file_path])
-        return  os.path.join(*clean)
+        clean = filter(None,[
+          self._path.intermediate_path,
+          self._path.dataset,
+          self._path.layer,
+          file_path
+        ])
+        return os.path.join(*clean)
 
     @retry
-    def put_file(self, file_path, content, content_type, compress):
+    def put_file(self, file_path, content, content_type, compress, cache_control=None):
         key = self.get_path_to_file(file_path)
         blob = self._bucket.blob( key )
         if compress:
             blob.content_encoding = "gzip"
+        if cache_control:
+            blob.cache_control = cache_control
         blob.upload_from_string(content, content_type)
 
     @retry
@@ -403,10 +551,10 @@ class GoogleCloudStorageInterface(object):
         key = self.get_path_to_file(file_path)
         blob = self._bucket.get_blob( key )
         if not blob:
-            return None, False
-        # blob handles the decompression in the case
-        # it is necessary
-        return blob.download_as_string(), False
+            return None, None # content, encoding
+
+        # blob handles the decompression so the encoding is None
+        return blob.download_as_string(), None # content, encoding
 
     def exists(self, file_path):
         key = self.get_path_to_file(file_path)
@@ -441,7 +589,7 @@ class GoogleCloudStorageInterface(object):
 
     def release_connection(self):
         global GC_POOL
-        GC_POOL.release_connection(self._client)
+        GC_POOL[self._path.bucket].release_connection(self._bucket)
 
 class S3Interface(object):
 
@@ -451,31 +599,30 @@ class S3Interface(object):
         self._conn = S3_POOL.get_connection()
 
     def get_path_to_file(self, file_path):
-        clean = filter(None,[self._path.dataset_path,
-                             self._path.dataset_name,
-                             self._path.layer_name,
+        clean = filter(None,[self._path.intermediate_path,
+                             self._path.dataset,
+                             self._path.layer,
                              file_path])
         return  os.path.join(*clean)
 
     @retry
-    def put_file(self, file_path, content, content_type, compress):
+    def put_file(self, file_path, content, content_type, compress, cache_control=None):
         key = self.get_path_to_file(file_path)
+
+        attrs = {
+            'Bucket': self._path.bucket,
+            'Body': content,
+            'Key': key,
+            'ContentType': (content_type or 'application/octet-stream'),
+        }
+
         if compress:
-            self._conn.put_object(
-                Bucket=self._path.bucket_name,
-                Body=content,
-                Key=key,
-                ContentType=(content_type or 'application/octet-stream'),
-                ContentEncoding='gzip',
-            )
-        else:
-            self._conn.put_object(
-                Bucket=self._path.bucket_name,
-                Body=content,
-                Key=key,
-                ContentType=(content_type or 'application/octet-stream'),
-            )
-            
+            attrs['ContentEncoding'] = 'gzip'
+        if cache_control:
+            attrs['CacheControl'] = cache_control
+
+        self._conn.put_object(**attrs)
+
     @retry
     def get_file(self, file_path):
         """
@@ -486,7 +633,7 @@ class S3Interface(object):
 
         try:
             resp = self._conn.get_object(
-                Bucket=self._path.bucket_name,
+                Bucket=self._path.bucket,
                 Key=self.get_path_to_file(file_path),
             )
 
@@ -494,10 +641,10 @@ class S3Interface(object):
             if 'ContentEncoding' in resp:
                 encoding = resp['ContentEncoding']
 
-            return resp['Body'].read(), encoding == "gzip"
+            return resp['Body'].read(), encoding
         except botocore.exceptions.ClientError as err: 
             if err.response['Error']['Code'] == 'NoSuchKey':
-                return None, False
+                return None, None
             else:
                 raise
 
@@ -505,7 +652,7 @@ class S3Interface(object):
         exists = True
         try:
             self._conn.head_object(
-                Bucket=self._path.bucket_name,
+                Bucket=self._path.bucket,
                 Key=self.get_path_to_file(file_path),
             )
         except botocore.exceptions.ClientError as e:
@@ -519,7 +666,7 @@ class S3Interface(object):
     @retry
     def delete_file(self, file_path):
         self._conn.delete_object(
-            Bucket=self._path.bucket_name,
+            Bucket=self._path.bucket,
             Key=self.get_path_to_file(file_path),
         )
 
@@ -534,20 +681,36 @@ class S3Interface(object):
 
         layer_path = self.get_path_to_file("")        
         path = os.path.join(layer_path, prefix)
+
         resp = self._conn.list_objects_v2(
-            Bucket=self._path.bucket_name,
+            Bucket=self._path.bucket,
             Prefix=path,
         )
 
-        if 'Contents' not in resp.keys():
-            resp['Contents'] = []
+        def iterate(resp):
+            if 'Contents' not in resp.keys():
+                resp['Contents'] = []
 
-        for item in resp['Contents']:
-            key = item['Key']
-            filename = key.replace(layer_path + '/', '')
-            if not flat and filename[-1] != '/':
-                yield filename
-            elif flat and '/' not in key.replace(path, ''):
+            for item in resp['Contents']:
+                key = item['Key']
+                filename = key.replace(layer_path + '/', '')
+                if not flat and filename[-1] != '/':
+                    yield filename
+                elif flat and '/' not in key.replace(path, ''):
+                    yield filename
+
+
+        for filename in iterate(resp):
+            yield filename
+
+        while resp['IsTruncated'] and resp['NextContinuationToken']:
+            resp = self._conn.list_objects_v2(
+                Bucket=self._path.bucket,
+                Prefix=path,
+                ContinuationToken=resp['NextContinuationToken'],
+            )
+
+            for filename in iterate(resp):
                 yield filename
 
     def release_connection(self):
